#include <QMessageBox>
#include <QProcess>
#include <QDir>

#include "applychangeswizard.h"
#include "ui_applychangeswizard.h"
#include "androidsdk.h"

ApplyChangesWizard::ApplyChangesWizard(const QString& sdkPath, const QList<Package>& uninstall, const QList<Package>& install, QWidget *parent) :
    QWizard(parent),
    ui(new Ui::ApplyChangesWizard),
    uninstall(uninstall),
    install(install),
    sdkPath(sdkPath),
    operationIdx(-1)
{
    ui->setupUi(this);

    setWindowFlags(windowFlags() & ~Qt::WindowCloseButtonHint);
}

ApplyChangesWizard::~ApplyChangesWizard()
{
    delete ui;
}

bool ApplyChangesWizard::isOperationInProgress()
{
    return (
        (currentPage() == ui->progressPage) &&
        !button(QWizard::FinishButton)->isEnabled()
    );
}

void ApplyChangesWizard::closeEvent(QCloseEvent* e)
{
    // Prevent closing via click

    if (isOperationInProgress()) {
       e->ignore();
    } else {
       QWizard::closeEvent(e);
    }
}

void ApplyChangesWizard::keyPressEvent(QKeyEvent *evt)
{
    // Prevent closing via Esc

    Q_UNUSED(evt);
}

void ApplyChangesWizard::on_ApplyChangesWizard_currentIdChanged(int id)
{
    Q_UNUSED(id);

    if (currentPage() == ui->overviewPage) {
        // Set up the overview page
        on_acceptCheck_stateChanged(0);
        ui->listWidget->clear();
        for (int i = 0; i < uninstall.size(); ++i) {
            ui->listWidget->addItem(tr("Uninstall '%1' v%2").arg(uninstall[i].Description).arg(uninstall[i].Version));
            ui->listWidget->item(ui->listWidget->count()-1)->setToolTip(tr("Component path: %1").arg(uninstall[i].Path));
        }
        for (int i = 0; i < install.size(); ++i) {
            ui->listWidget->addItem(tr("Install '%1' v%2").arg(install[i].Description).arg(install[i].Version));
            ui->listWidget->item(ui->listWidget->count()-1)->setToolTip(tr("Component path: %1").arg(install[i].Path));
        }

        if (install.size() == 0 && uninstall.size() == 0) {
            QMessageBox qmb;
            qmb.setIcon(QMessageBox::Critical);
            qmb.setText(tr("No changes selected."));
            qmb.exec();
            reject();
            return;
        }

    } else if (currentPage() == ui->licensePage) {
        // License page.
        // Nothing special happens here

    } else if (currentPage() == ui->progressPage) {
        // Start processing
        button(QWizard::FinishButton)->setEnabled(false);
        button(QWizard::CancelButton)->setEnabled(false);
        button(QWizard::BackButton)->setEnabled(false);
        ui->progressBar->setMaximum(install.size() + uninstall.size());

        nextOperation();
    }
}

void ApplyChangesWizard::nextOperation()
{
    operationIdx++;
    ui->progressBar->setValue(operationIdx);

    if (operationIdx < uninstall.size()) {
        // Perform an uninstallation
        QString p = uninstall[operationIdx].Path;
        ui->outputArea->appendPlainText(tr("Uninstalling '%1'...\n").arg(p));
        sdkCommand(QStringList{"--uninstall", p});

    } else if (operationIdx < (install.size() + uninstall.size())) {
        QString p = install[operationIdx - uninstall.size()].Path;
        ui->outputArea->appendPlainText(tr("Installing '%1'...\n").arg(p));
        sdkCommand(QStringList{p});

    } else {
        // Reached the end
        button(QWizard::FinishButton)->setEnabled(true);
        ui->progressPage->setSubTitle(tr("All operations completed."));
    }
}

void ApplyChangesWizard::sdkCommand(QStringList args)
{
    QProcess *p = new QProcess(this);

    connect(p, &QProcess::errorOccurred, this, [=](QProcess::ProcessError pe) {
        Q_UNUSED(pe);
        processFailure(tr("Failed to launch sdkmanager: %1").arg(p->errorString()));
        p->close();
        p->deleteLater();
    });

    connect(p, &QProcess::readyReadStandardOutput, this, [=]() {
        ui->outputArea->appendPlainText(p->readAllStandardOutput());
    });

    connect(p, &QProcess::readyReadStandardError, this, [=]() {
        ui->outputArea->appendPlainText(p->readAllStandardError());
    });

    connect(p, &QProcess::started, this, [=](){
        // Automatically accept license prompts
        p->write("y\n");
        p->closeWriteChannel();
    });

    connect(p, static_cast<void (QProcess::*)(int)>(&QProcess::finished), this, [=](int exitCode) {
        ui->outputArea->appendPlainText(p->readAllStandardOutput());
        ui->outputArea->appendPlainText(p->readAllStandardError());
        p->deleteLater();

        if (exitCode == 0) {
            nextOperation();
        } else {
            processFailure(tr("Process exit code %1").arg(exitCode));
        }
    });

    p->setWorkingDirectory(sdkPath);
    p->start(QDir(sdkPath).absoluteFilePath(SDKMANAGER_REL_PATH), args);

}

void ApplyChangesWizard::processFailure(QString error)
{
    ui->outputArea->appendPlainText(tr("Error performing operation."));
    ui->outputArea->appendPlainText(error);

    QMessageBox qmb;
    qmb.setIcon(QMessageBox::Critical);
    qmb.setText(error);
    qmb.exec();

    button(QWizard::FinishButton)->setEnabled(true);
}

void ApplyChangesWizard::on_acceptCheck_stateChanged(int arg1)
{
    Q_UNUSED(arg1);
    if (currentPage() != ui->overviewPage) {
        return; // irrelevant
    }

    button(QWizard::NextButton)->setEnabled( ui->acceptCheck->isChecked() );
}
