#include <QFileDialog>
#include <QLabel>
#include <QProcess>
#include <QRegularExpression>
#include <QMessageBox>
#include <QDesktopServices>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "package.h"
#include "unify.h"
#include "androidsdk.h"
#include "applychangeswizard.h"

static const int UISTATE_NO_SDK_PATH           = 1;
static const int UISTATE_SCANNING_IN_PROGRESS  = 2;
static const int UISTATE_SCAN_COMPLETED        = 3;

static const int CHANNEL_STABLE = 0;
static const int CHANNEL_BETA   = 1;
static const int CHANNEL_DEV    = 2;
static const int CHANNEL_CANARY = 3;

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    channel(0)
{
    ui->setupUi(this);
    QApplication::setApplicationDisplayName(windowTitle());

    statusLabel = new QLabel();
    ui->statusBar->addWidget(statusLabel);

    updateUIState(UISTATE_NO_SDK_PATH);

#ifdef QT_DEBUG
    setSDKPath("/opt/android-sdk/");
#else
    setSDKPath(QString::fromUtf8(qgetenv("ANDROID_SDK_HOME")));
#endif
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::updateUIState(int state)
{
    ui->packageTree        ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionRescan       ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionApply_changes->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionCollapse_all ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionExpand_all   ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionInclude_obsolete_packages->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionStable->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionBeta  ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionDev   ->setEnabled(state == UISTATE_SCAN_COMPLETED);
    ui->actionCanary->setEnabled(state == UISTATE_SCAN_COMPLETED);

    switch(state) {

    case UISTATE_NO_SDK_PATH: {
        ui->actionChoose_SDK->setEnabled(true);
    } break;

    case UISTATE_SCANNING_IN_PROGRESS: {
        ui->actionChoose_SDK->setEnabled(false);
    } break;

    case UISTATE_SCAN_COMPLETED: {
        ui->actionChoose_SDK->setEnabled(true);
    } break;
    }
}

void MainWindow::setSDKPath(QString path) {
    currentSDKPath = path;

    if (path.size() > 0) {

        // Validate android-sdk exists at selected path
        if (QDir(path).exists() && QFile(QDir(path).absoluteFilePath(SDKMANAGER_REL_PATH)).exists()) {
            rescan();

        } else {
            statusLabel->setText(tr("Selected directory does not contain an Android SDK."));
            updateUIState(UISTATE_NO_SDK_PATH);
        }

    } else {
        statusLabel->setText(tr("No Android SDK selected."));
        updateUIState(UISTATE_NO_SDK_PATH);

    }
}

void MainWindow::rescan()
{
    if (currentSDKPath.size() == 0) {
        return; // can't rescan
    }

    updateUIState(UISTATE_SCANNING_IN_PROGRESS);

    statusLabel->setText(tr("Scanning..."));

    QProcess *p = new QProcess(this);
    connect(p, &QProcess::errorOccurred, this, [=](QProcess::ProcessError pe) {
        Q_UNUSED(pe);
        statusLabel->setText(tr("Failed to launch sdkmanager: %1").arg(p->errorString()));
        ui->actionRescan->setEnabled(true);

        p->close();
        p->deleteLater();
    });
    connect(p, static_cast<void (QProcess::*)(int)>(&QProcess::finished), this, [=](int exitCode) {
        if (exitCode == 0) {
            refreshListFromOutput(p->readAllStandardOutput());
        } else {
            statusLabel->setText(tr("Failed to enumerate packages."));
            ui->actionRescan->setEnabled(true);
        }
        p->deleteLater();
    });

    QStringList params{"--list", "--verbose"};
    if (ui->actionInclude_obsolete_packages->isChecked()) {
        params.append("--include_obsolete");
    }
    params.append(QString("--channel=%1").arg(channel));

    p->setWorkingDirectory(currentSDKPath);
    p->start(QDir(currentSDKPath).absoluteFilePath(SDKMANAGER_REL_PATH), params);
}

void MainWindow::refreshListFromOutput(QByteArray o)
{
    ui->packageTree->clear();

    lastScan = Package::Parse(o);

    for (int i = 0; i < lastScan.size(); ++i) {
        addPackage(i, lastScan[i]);
    }

    statusLabel->setText(tr("Scan complete."));

    ui->packageTree->expandAll();
    ui->packageTree->resizeColumnToContents(0);
    ui->packageTree->resizeColumnToContents(1);

    ui->packageTree->sortItems(0, Qt::AscendingOrder);

    updateUIState(UISTATE_SCAN_COMPLETED);
}

static const int PathRole       = Qt::UserRole + 1;
static const int FullPathRole   = Qt::UserRole + 2;
static const int PackageIdxRole = Qt::UserRole + 3;

void MainWindow::addPackage(int packageIdx, const Package& p)
{
    Unify *parent = new UnifyWidget(ui->packageTree);

    QStringList parts = p.Path.split(';');
    QString lastPart = parts[parts.size() - 1];
    for(int i = 0; i < parts.size() - 1; ++i) {

        QTreeWidgetItem *child = nullptr;
        for (int j = 0; j < parent->childCount(); ++j) {
            QString path = parent->childAt(j)->data(0, PathRole).toString();
            if (path == parts[i]) { // found
                child = parent->childAt(j);
                break; // inner loop
            }
        }

        if (child == nullptr) {
            // There was no existing subcomponent with this path
            // Create a new one
            auto x = new QTreeWidgetItem();
            x->setData(0, PathRole, parts[i]);
            x->setText(0, parts[i]);
            x->setIcon(0, iconProvider.icon(QFileIconProvider::Folder));
            parent->addChild(x); // takes ownership
            delete parent;
            parent = new UnifyItem(x);
        } else {
            // Matched an existing subcomponent, continue
            delete parent;
            parent = new UnifyItem(child);
        }

    }

    QTreeWidgetItem *itm = new QTreeWidgetItem();
    itm->setData(0, Qt::ToolTipRole, tr("Component path: %1").arg(p.Path));
    itm->setData(0, FullPathRole, p.Path);
    itm->setData(0, PackageIdxRole, packageIdx);
    itm->setText(0, lastPart);
    itm->setText(1, p.Version);
    itm->setText(2, p.Description);
    itm->setIcon(0, iconProvider.icon(QFileIconProvider::File));

    if (p.CurrentlyInstalled) {
        itm->setCheckState(0, Qt::Checked);
    } else {
        itm->setCheckState(0, Qt::Unchecked);
    }

    // Add to treewidget
    parent->addChild(itm); // takes ownership

    delete parent; // free allocated virtual interface
}

void MainWindow::visitPackages(QTreeWidgetItem* root, std::function<void(QTreeWidgetItem*)> cb)
{
    Unify *u;
    if (root == nullptr) {
        u = new UnifyWidget(ui->packageTree);
    } else {
        u = new UnifyItem(root);
    }

    for (int i = 0; i < u->childCount(); ++i) {
        QString pkg = u->childAt(i)->data(0, FullPathRole).toString();
        if (pkg.size() > 0) {
            cb(u->childAt(i));
        }
        if (u->childAt(i)->childCount() > 0) {
            visitPackages(u->childAt(i), cb);
        }
    }

    delete u;
}

void MainWindow::on_actionChoose_SDK_triggered()
{
    QString sdkDir = QFileDialog::getExistingDirectory(this, tr("Select Android SDK directory..."));

    if (sdkDir.size() == 0) {
        return;
    }

    // TODO somehow validate that this directory is real
    currentSDKPath = sdkDir;
    rescan();
}

void MainWindow::on_actionRescan_triggered()
{
    rescan();
}

void MainWindow::on_actionE_xit_triggered()
{
    close();
}

void MainWindow::on_actionExpand_all_triggered()
{
    ui->packageTree->expandAll();
}

void MainWindow::on_actionCollapse_all_triggered()
{
    ui->packageTree->collapseAll();
}

void MainWindow::on_actionApply_changes_triggered()
{
    // Walk the UI tree, determine install/uninstall by comparing check states
    QList<Package> toInstall;
    QList<Package> toUninstall;
    visitPackages(nullptr, [&](QTreeWidgetItem *itm) {
        int idx = itm->data(0, PackageIdxRole).toInt();
        bool selected = (itm->checkState(0) == Qt::Checked);
        if (lastScan[idx].CurrentlyInstalled && !selected) {
            toUninstall.push_back(lastScan[idx]);
        } else if (!lastScan[idx].CurrentlyInstalled && selected) {
            toInstall.push_back(lastScan[idx]);
        }
    });

    auto *acw = new ApplyChangesWizard(currentSDKPath, toUninstall, toInstall, this);
    acw->setWindowModality(Qt::ApplicationModal);
    connect(acw, &QWizard::finished, this, [=](int result) {
        if (result == QWizard::Accepted) {
            rescan();
        }
        acw->deleteLater();
    });
    acw->show();
}

void MainWindow::on_actionAbout_qsdkmanager_triggered()
{
    QMessageBox qmb;
    qmb.setText(QApplication::applicationDisplayName() + "\n" + tr("Version: %1").arg(QApplication::applicationVersion()));
    qmb.setIconPixmap(QIcon(":/rsrc/ic_launcher.png").pixmap(64, 64));
    qmb.exec();
}

void MainWindow::on_actionAbout_Qt_triggered()
{
    QApplication::aboutQt();
}

void MainWindow::on_actionCurrent_triggered()
{
    visitPackages(nullptr, [=](QTreeWidgetItem *itm) {
        int idx = itm->data(0, PackageIdxRole).toInt();
        itm->setCheckState(0, lastScan[idx].CurrentlyInstalled ? Qt::Checked : Qt::Unchecked);
    });
}

void MainWindow::on_actionAll_triggered()
{
    visitPackages(nullptr, [=](QTreeWidgetItem * itm) {
        itm->setCheckState(0, Qt::Checked);
    });
}

void MainWindow::on_actionNone_triggered()
{
    visitPackages(nullptr, [=](QTreeWidgetItem * itm) {
        itm->setCheckState(0, Qt::Unchecked);
    });
}

void MainWindow::on_actionInclude_obsolete_packages_triggered()
{
    // Qt automatically toggles the checkbox for us
    rescan();
}

void MainWindow::setChannel(int c)
{
    channel = c;
    ui->actionStable->setChecked(c == CHANNEL_STABLE);
    ui->actionBeta  ->setChecked(c == CHANNEL_BETA);
    ui->actionDev   ->setChecked(c == CHANNEL_DEV);
    ui->actionCanary->setChecked(c == CHANNEL_CANARY);
    rescan();
}

void MainWindow::on_actionStable_triggered()
{
    setChannel(CHANNEL_STABLE);
}

void MainWindow::on_actionBeta_triggered()
{
    setChannel(CHANNEL_BETA);
}

void MainWindow::on_actionDev_triggered()
{
    setChannel(CHANNEL_DEV);
}

void MainWindow::on_actionCanary_triggered()
{
    setChannel(CHANNEL_CANARY);
}

void MainWindow::on_actionHomepage_triggered()
{
    QDesktopServices::openUrl(QUrl("https://code.ivysaur.me/qsdkmanager/"));
}
